from contextlib import contextmanager
import pathlib
import sys
from typing import Union, List
from .import_hook import PyFinder, PyHTTPFinder

# Singleton instance of PyFinder
pyfinder: PyFinder = None


def _update_syspath(path: str):
    """
    Append `path` to sys.path so that files in path can be imported
    """
    path = pathlib.Path(path).resolve().as_posix()
    if path not in sys.path:
        sys.path.append(path)


def register_hook(
    base_url: str,
    download_path: str = "",
    modules: List[str] = None,
    update_syspath: bool = True,
):
    """
    Register PyFinder import hook to sys.meta_path
    """
    global pyfinder
    if pyfinder is not None and pyfinder._registered():
        raise RuntimeError(
            "import hook is already registered, if you want to register a new hook, unregister the existing hook with unregister_hook() first"
        )

    pyfinder = PyHTTPFinder(base_url, download_path, modules)
    pyfinder.register()
    if update_syspath:
        _update_syspath(download_path)

    return pyfinder


def unregister_hook():
    """
    Unregister PyFinder import hook from sys.meta_path
    """
    global pyfinder

    if pyfinder is not None:
        pyfinder.unregister()
        pyfinder = None


def add_module(module: Union[str, List[str]]):
    """
    Add new module(s) to the whitelist which PyFinder import hook can import
    """
    global pyfinder

    if pyfinder is None or (not pyfinder._registered()):
        raise RuntimeError("import hook is not registered")

    pyfinder.add_module(module)


def available_modules():
    """
    Get the list of whitelisted modules which PyFinder import hook can import
    """
    global pyfinder
    if pyfinder is None or (not pyfinder._registered()):
        raise RuntimeError("import hook is not registered")

    return pyfinder.available_modules()



import tempfile
import sys

import pytest

import pyodide_importer

TEST_MODULE_URL = (
    "https://raw.githubusercontent.com/ryanking13/pyodide-importer/main/test_modules/"
)


@pytest.fixture(scope="function")
def hook():
    tempdir = tempfile.TemporaryDirectory()
    pyfinder = pyodide_importer.register_hook(
        base_url=TEST_MODULE_URL,
        download_path=tempdir.name,
        update_syspath=True,
    )

    yield pyfinder

    pyodide_importer.unregister_hook()
    tempdir.cleanup()


@pytest.fixture(autouse=True)
def cleanup_imports():
    yield

    # Cleanup imported modules
    cleanup_targets = []
    for module in sys.modules:
        if module.startswith(("file_module", "regular_module")):
            cleanup_targets.append(module)

    for module in cleanup_targets:
        sys.modules.pop(module)


def test_file_module(hook):
    expected_response = "hello from file_module"

    import file_module

    assert file_module.hello() == expected_response


def test_regular_module_init(hook):
    expected_response = "hello from regular_module"

    import regular_module

    assert regular_module.hello() == expected_response


def test_regular_module_submodule(hook):
    expected_response = "hello from regular_module.submodule"

    import regular_module.submodule

    assert regular_module.submodule.hello() == expected_response


def test_regular_module_submodule2(hook):
    expected_response = "hello from regular_module.submodule2"

    import regular_module

    assert regular_module.submodule2_hello() == expected_response


def test_module_whitelist(hook):
    pyodide_importer.add_module("file_module")

    import file_module

    with pytest.raises(ModuleNotFoundError):
        import regular_module


def test_nohook():
    with pytest.raises(ModuleNotFoundError):
        import file_module


def test_unregister_hook(hook):
    pyodide_importer.unregister_hook()
    with pytest.raises(ModuleNotFoundError):
        import file_module


def test_available_modules(hook):
    hook.add_module(["module1", "module2"])

    modules = hook.available_modules()
    assert "module1" in modules
    assert "notamodule" not in modules


def test_contextmanager(hook):
    with hook:
        import file_module

    with pytest.raises(ModuleNotFoundError):
        import regular_module



from importlib.abc import MetaPathFinder
from _frozen_importlib_external import PathFinder
import sys
import pathlib
import warnings
from typing import List, Any, Union, Iterable

PYODIDE = True

try:
    from js import XMLHttpRequest
except:
    warnings.warn("Pyodide not found, running in native mode")
    import requests

    PYODIDE = False


class PyFinder(MetaPathFinder):
    """Meta path finder for Pyodide which supports importing external Python files"""

    def __init__(
        self,
        import_path: Union[str, List[str]] = "",
        download_path: str = "",
        modules: Iterable[str] = None,
    ):
        self.import_path = self._prepare_import_path(import_path)
        self.download_path = self._to_abspath(download_path)
        self.modules = set(modules) if modules is not None else None

    def register(self):
        """
        Register the PyFinder to sys.meta_path
        """
        if not self._registered():
            sys.meta_path.append(self)

    def unregister(self):
        """
        Unregister the PyFinder from sys.meta_path
        """
        if self._registered():
            sys.meta_path.remove(self)

    def _registered(self):
        """
        Check whether the PyFinder is registered in sys.meta_path
        """
        return self in sys.meta_path

    def __enter__(self):
        self.register()

    def __exit__(self, type, value, traceback):
        self.unregister()

    def _prepare_import_path(self, paths: Union[str, List[str]]):
        if isinstance(paths, str):
            return [paths.rstrip("/") + "/"]
        elif isinstance(paths, list):
            return [path.rstrip("/") + "/" for path in paths]

    def _to_abspath(self, path: str):
        return pathlib.Path(path).resolve()

    def add_module(self, module: Union[str, List[str]]):
        """
        Add new module(s) to the whitelist which PyFinder import hook can import
        """
        if self.modules is None:
            self.modules = set()

        if isinstance(module, str):
            self.modules.add(module)
        else:
            self.modules.update(module)

    def available_modules(self):
        """
        Get the list of whitelisted modules which PyFinder import hook can import
        """
        return self.modules

    @staticmethod
    def invalidate_caches():
        """Call the invalidate_caches() method on all path entry finders
        stored in sys.path_importer_caches (where implemented)."""
        for name, finder in list(sys.path_importer_cache.items()):
            if finder is None:
                del sys.path_importer_cache[name]
            elif hasattr(finder, "invalidate_caches"):
                finder.invalidate_caches()

    def find_spec(self, fullname: str, path: List[str], target: Any = None):
        # Should find python packages according to PEP-420.
        raise NotImplementedError


class PyHTTPFinder(PyFinder):
    """Meta path finder for importing python files through HTTP(s)"""

    def __init__(
        self,
        import_path: Union[str, List[str]] = "",
        download_path: str = "",
        modules: Iterable[str] = None,
    ):
        super().__init__(import_path, download_path, modules)

    def _get(self, url: str):
        """
        Download python script through HTTP(s)
        """
        if PYODIDE:
            req = XMLHttpRequest.new()
            req.open("GET", url, False)
            req.send(None)
            return (req.status, req.responseText)
        else:
            # This enables testing this module outside of Pyodide
            resp = requests.get(url)
            return (resp.status_code, resp.text)

    def find_spec(self, fullname: str, path: List[str], target: Any = None):
        # Search order:
        # 1. <directory>/foo/__init__.py
        # 2. <directory>/foo.py
        # (-) Not supported: binary formats (pyc, pyd, so)
        # (-) Not supported: Namespace packages without __init__.py

        # module1.module2 ==> module1/module2
        modules = fullname.split(".")
        if self.modules is not None and modules[0] not in self.modules:
            return None

        fullname_aspath = fullname.replace(".", "/")

        import_subpaths = [
            f"{fullname_aspath}/__init__.py",  # Regular package
            f"{fullname_aspath}.py",  # Module package
        ]

        for base_path in self.import_path:
            for import_subpath in import_subpaths:
                status_code, content = self._get(f"{base_path}{import_subpath}")
                if status_code != 200:
                    continue

                package_path = self.download_path / import_subpath
                package_path.parent.mkdir(parents=True, exist_ok=True)
                with package_path.open("w") as f:
                    f.write(content)

                # We need to invalidate caches since PathFinder have already failed once and it is cached.
                self.invalidate_caches()
                return PathFinder.find_spec(fullname, path, target)

        return None
